/* GStreamer
 * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
 *                    2000 Wim Taymans <wtay@chello.be>
 *
 * cothreads.h: Header for cothreading routines
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __COTHREADS_H__
#define __COTHREADS_H__

#include <glib.h>
#include <setjmp.h>

#ifdef HAVE_ATOMIC_H
#include <asm/atomic.h>
#endif

#undef COTHREAD_ATOMIC

#ifndef CURRENT_STACK_FRAME
#define CURRENT_STACK_FRAME  ({ char __csf; &__csf; })
#endif /* CURRENT_STACK_FRAME */

typedef struct _cothread_state		cothread_state;
typedef struct _cothread_context	cothread_context;

typedef int (*cothread_func) (int argc,char **argv);

#define COTHREAD_STARTED	0x01
#define COTHREAD_DESTROYED	0x02

/* moved the COTHREAD_ATOMIC dependent stuff here instead of littered around
 * the code as previous */

/* _lock_t type */
#ifdef COTHREAD_ATOMIC
typedef atomic_t _lock_t;
#else
typedef GMutex * _lock_t;
#endif

/* _lock_new macro */
#ifdef COTHREAD_ATOMIC
#  define _lock_new(x)  atomic_set (x, 0);
#else
#  define _lock_new(x)  x = g_mutex_new ();
#endif
    
/* _lock_free macro */
#ifdef COTHREAD_ATOMIC
#  define _lock_free(x) atomic_set (x, 0);
#else
#  define _lock_free(x) g_mutex_free (x);
#endif

/* _lock_lock macro */
#ifdef COTHREAD_ATOMIC
#  define _lock_lock(x) atomic_set (x, 1);
#else
#  define _lock_lock(x) if (x) { g_mutex_lock (x); }
#endif

/* _lock_unlock macro */
#ifdef COTHREAD_ATOMIC
#  define _lock_unlock(x) atomic_set (x, 0);
#else
#  define _lock_unlock(x) if (x) { g_mutex_unlock (x); }
#endif

/* _lock_trylock macro */
#ifdef COTHREAD_ATOMIC
#  define _lock_trylock(x) \
  if (atomic_read (x) == 1) \
    return FALSE; \
  else { \
    _lock_lock(x); \
    return TRUE; }
#else
#  define _lock_trylock(x) return g_mutex_trylock (x);
#endif
    
struct _cothread_state {
  cothread_context *ctx;
  int threadnum;
  gpointer priv;

  cothread_func func;
  int argc;
  char **argv;

  int flags;
  void *sp;
  jmp_buf jmp;
  /* is this needed any more? */
  void *top_sp;
  void *pc;

  int magic_number;
  _lock_t lock;
};


cothread_context*		cothread_context_init   	(void);
void				cothread_context_free		(cothread_context *ctx);
void				cothread_context_set_data	(cothread_state *thread, 
								 gchar *key, gpointer data);
gpointer			cothread_context_get_data	(cothread_state *thread, gchar *key);

cothread_state*			cothread_create			(cothread_context *ctx);
void				cothread_free			(cothread_state *thread);
void				cothread_setfunc		(cothread_state *thread, cothread_func func,
						        	 int argc, char **argv);
void				cothread_stop			(cothread_state *thread);

void				cothread_switch			(cothread_state *thread);
gboolean			cothread_stackquery		(void **stack, glong* stacksize);
void				cothread_set_private		(cothread_state *thread, 
								 gpointer data);
gpointer			cothread_get_private		(cothread_state *thread);

void				cothread_lock			(cothread_state *thread);
gboolean			cothread_trylock		(cothread_state *thread);
void				cothread_unlock			(cothread_state *thread);

cothread_state*			cothread_main			(cothread_context *ctx);
cothread_state*			cothread_current_main		(void);
cothread_state*			cothread_current		(void);

#endif /* __COTHREAD_H__ */
